/*
  * JBoss, Home of Professional Open Source
  * Copyright 2005, JBoss Inc., and individual contributors as indicated
  * by the @authors tag. See the copyright.txt in the distribution for a
  * full listing of individual contributors.
  *
  * This is free software; you can redistribute it and/or modify it
  * under the terms of the GNU Lesser General Public License as
  * published by the Free Software Foundation; either version 2.1 of
  * the License, or (at your option) any later version.
  *
  * This software is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this software; if not, write to the Free
  * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
  * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
  */
package org.jboss.jms.server.endpoint;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.jms.IllegalStateException;
import javax.jms.InvalidDestinationException;
import javax.jms.JMSException;

import org.jboss.aop.AspectManager;
import org.jboss.jms.client.delegate.ClientBrowserDelegate;
import org.jboss.jms.client.delegate.ClientConsumerDelegate;
import org.jboss.jms.delegate.BrowserDelegate;
import org.jboss.jms.delegate.ConsumerDelegate;
import org.jboss.jms.destination.JBossDestination;
import org.jboss.jms.destination.JBossQueue;
import org.jboss.jms.destination.JBossTopic;
import org.jboss.jms.message.JBossMessage;
import org.jboss.jms.server.ServerPeer;
import org.jboss.jms.server.endpoint.advised.BrowserAdvised;
import org.jboss.jms.server.endpoint.advised.ConsumerAdvised;
import org.jboss.jms.server.plugin.contract.ChannelMapper;
import org.jboss.jms.server.remoting.JMSDispatcher;
import org.jboss.jms.server.subscription.DurableSubscription;
import org.jboss.jms.server.subscription.Subscription;
import org.jboss.jms.tx.AckInfo;
import org.jboss.jms.util.ExceptionUtil;
import org.jboss.logging.Logger;
import org.jboss.messaging.core.Channel;
import org.jboss.messaging.core.local.CoreDestination;
import org.jboss.messaging.core.local.CoreSubscription;
import org.jboss.messaging.core.local.Queue;
import org.jboss.messaging.core.memory.MemoryManager;
import org.jboss.messaging.core.plugin.contract.MessageStore;
import org.jboss.messaging.core.plugin.contract.PersistenceManager;
import org.jboss.messaging.core.tx.Transaction;
import org.jboss.messaging.core.tx.TransactionRepository;

/**
 * Concrete implementation of SessionEndpoint.
 * 
 * @author <a href="mailto:ovidiu@jboss.org">Ovidiu Feodorov</a>
 * @author <a href="mailto:tim.fox@jboss.com">Tim Fox</a>
 * @version <tt>$Revision: 2637 $</tt>
 *
 * $Id: ServerSessionEndpoint.java 2637 2007-05-04 09:34:46Z timfox $
 */
public class ServerSessionEndpoint implements SessionEndpoint
{
   // Constants -----------------------------------------------------

   private static final Logger log = Logger.getLogger(ServerSessionEndpoint.class);

   // Static --------------------------------------------------------

   // Attributes ----------------------------------------------------
   
   private boolean trace = log.isTraceEnabled();

   private int sessionID;
   
   private boolean closed;

   private ServerConnectionEndpoint connectionEndpoint;

   private Map consumers;
   private Map browsers;

   private ChannelMapper cm;
   private PersistenceManager pm;
   private MessageStore ms;
   private MemoryManager mm;
   private CoreDestination dlq;
   private TransactionRepository tr;
   private int maxDeliveryAttempts;
   
   // Constructors --------------------------------------------------

   protected ServerSessionEndpoint(int sessionID, ServerConnectionEndpoint connectionEndpoint)
      throws Exception
   {
      this.sessionID = sessionID;
      
      this.connectionEndpoint = connectionEndpoint;

      ServerPeer sp = connectionEndpoint.getServerPeer();

      cm = sp.getChannelMapperDelegate();
      pm = sp.getPersistenceManagerDelegate();
      ms = sp.getMessageStoreDelegate();
      mm = sp.getMemoryManager();

      consumers = new HashMap();
		browsers = new HashMap();  
      
      dlq = sp.getDLQ();
      tr = sp.getTxRepository();
      maxDeliveryAttempts = sp.getMaxDeliveryAttempts();
   }
   
   // SessionDelegate implementation --------------------------------

	public ConsumerDelegate createConsumerDelegate(JBossDestination jmsDestination,
                                                  String selector,
                                                  boolean noLocal,
                                                  String subscriptionName,
                                                  boolean isCC) throws JMSException
   {
      try
      {
         if (closed)
         {
            throw new IllegalStateException("Session is closed");
         }
         
         if ("".equals(selector))
         {
            selector = null;
         }
         
         log.debug("creating consumer for " + jmsDestination + ", selector " + selector + ", " + (noLocal ? "noLocal, " : "") + "subscription " + subscriptionName);
   
         if (jmsDestination.isTemporary())
         {
            // Can only create a consumer for a temporary destination on the same connection
            // that created it
            if (!connectionEndpoint.hasTemporaryDestination(jmsDestination))
            {
               String msg = "Cannot create a message consumer on a different connection " +
                            "to that which created the temporary destination";
               throw new IllegalStateException(msg);
            }
         }
         
         CoreDestination coreDestination = cm.getCoreDestination(jmsDestination);
         if (coreDestination == null)
         {
            throw new InvalidDestinationException("No such destination: " + jmsDestination);
         }
             
         int consumerID = connectionEndpoint.getServerPeer().getNextObjectID();
        
         Subscription subscription = null;
   
         if (jmsDestination.isTopic())
         {
            if (subscriptionName == null)
            {
               // non-durable subscription
               if (log.isTraceEnabled()) { log.trace("creating new non-durable subscription on " + coreDestination); }
               subscription =
                  cm.createSubscription(jmsDestination.getName(), selector, noLocal, ms, pm, mm);
            }
            else
            {
               if (jmsDestination.isTemporary())
               {
                  throw new InvalidDestinationException("Cannot create a durable subscription on a temporary topic");
               }
               
               // we have a durable subscription, look it up
               String clientID = connectionEndpoint.getClientID();
               if (clientID == null)
               {
                  throw new JMSException("Cannot create durable subscriber without a valid client ID");
               }
   
               subscription = cm.getDurableSubscription(clientID, subscriptionName, ms, pm, mm);
   
               if (subscription == null)
               {
                  if (trace) { log.trace("creating new durable subscription on " + coreDestination); }
                  subscription = cm.createDurableSubscription(jmsDestination.getName(),
                                                              clientID,
                                                              subscriptionName,
                                                              selector,
                                                              noLocal,
                                                              ms,
                                                              pm,
                                                              mm);
               }
               else
               {
                  if (trace) { log.trace("subscription " + subscriptionName + " already exists"); }
                  
                  //Can't create subscriber if there is already an active subscriber
                  if (subscription.iterator().hasNext())
                  {
                  	throw new IllegalStateException("Cannot create a subscriber on the durable subscription since it already has subscriber(s)");

                  }
   
                  // From javax.jms.Session Javadoc (and also JMS 1.1 6.11.1):
                  // A client can change an existing durable subscription by creating a durable
                  // TopicSubscriber with the same name and a new topic and/or message selector.
                  // Changing a durable subscriber is equivalent to unsubscribing (deleting) the old
                  // one and creating a new one.
   
                  boolean selectorChanged =
                     (selector == null && subscription.getSelector() != null) ||
                     (subscription.getSelector() == null && selector != null) ||
                     (subscription.getSelector() != null && selector != null &&
                     !subscription.getSelector().equals(selector));
                  
                  if (trace) { log.trace("selector " + (selectorChanged ? "has" : "has NOT") + " changed"); }
   
                  boolean topicChanged =  subscription.getTopic().getId() != coreDestination.getId();
                  
                  if (log.isTraceEnabled()) { log.trace("topic " + (topicChanged ? "has" : "has NOT") + " changed"); }
                  
                  boolean noLocalChanged = noLocal != subscription.isNoLocal();
   
                  if (selectorChanged || topicChanged || noLocalChanged)
                  {
                     if (trace) { log.trace("topic or selector or noLocal changed so deleting old subscription"); }
   
                     boolean removed =
                        cm.removeDurableSubscription(connectionEndpoint.getClientID(), subscriptionName);
   
                     if (!removed)
                     {
                        throw new InvalidDestinationException("Cannot find durable subscription " +
                                                              subscriptionName + " to unsubscribe");
                     }
   
                     //Remove data for the durable sub
                     ((DurableSubscription)subscription).unsubscribe();
   
                     // create a fresh new subscription
                     subscription = cm.createDurableSubscription(jmsDestination.getName(),
                                                                 clientID,
                                                                 subscriptionName,
                                                                 selector,
                                                                 noLocal,
                                                                 ms,
                                                                 pm,
                                                                 mm);
                  }               
               }
            }
         }
         
         int prefetchSize = connectionEndpoint.getPrefetchSize();
         
         ServerConsumerEndpoint ep =
            new ServerConsumerEndpoint(consumerID,
                                       subscription == null ? (Channel)coreDestination : subscription,
                                       this, selector, noLocal, jmsDestination, prefetchSize, dlq);
          
         ConsumerAdvised advised;
         
         // Need to synchronize due to http://jira.jboss.com/jira/browse/JBMESSAGING-797               
         synchronized (AspectManager.instance())
         {
            advised = new ConsumerAdvised(ep);
         }                  
         
         JMSDispatcher.instance.registerTarget(new Integer(consumerID), advised);            
         
         ClientConsumerDelegate stub = new ClientConsumerDelegate(consumerID, prefetchSize, maxDeliveryAttempts);
         
         if (subscription != null)
         {
            subscription.connect();
         }
               
         putConsumerEndpoint(consumerID, ep); // caching consumer locally
         connectionEndpoint.getServerPeer().putConsumerEndpoint(consumerID, ep); // cachin consumer in server peer
         
         log.debug("created and registered " + ep);
   
         return stub;
      }
      catch (Throwable t)
      {
         throw ExceptionUtil.handleJMSInvocation(t, this + " createConsumerDelegate");
      }
   }
	
	public BrowserDelegate createBrowserDelegate(JBossDestination jmsDestination, String messageSelector)
	   throws JMSException
	{
      try
      {
   	   if (closed)
   	   {
   	      throw new IllegalStateException("Session is closed");
   	   }
   	   
   	   if (jmsDestination == null)
   	   {
   	      throw new InvalidDestinationException("null destination");
   	   }
   	   
   	   CoreDestination destination = cm.getCoreDestination(jmsDestination);
   	   
   	   if (destination == null)
   	   {
   	      throw new InvalidDestinationException("No such destination: " + jmsDestination);
   	   }
   	   
   	   if (!(destination instanceof Queue))
   	   {
   	      throw new IllegalStateException("Cannot browse a topic");
   	   }
   	   
   	   int browserID = connectionEndpoint.getServerPeer().getNextObjectID();
   	   
   	   ServerBrowserEndpoint ep =
   	      new ServerBrowserEndpoint(this, browserID, (Channel)destination, messageSelector);
   	   
   	   putBrowserDelegate(browserID, ep);
         
         BrowserAdvised advised;
         
         // Need to synchronize due to http://jira.jboss.com/jira/browse/JBMESSAGING-797               
         synchronized (AspectManager.instance())
         { 
            advised = new BrowserAdvised(ep);
         }
   	   
         JMSDispatcher.instance.registerTarget(new Integer(browserID), advised);
   	   
   	   ClientBrowserDelegate stub = new ClientBrowserDelegate(browserID);
   	   
         log.debug("created and registered " + ep);
   
   	   return stub;
      }
      catch (Throwable t)
      {
         throw ExceptionUtil.handleJMSInvocation(t, this + " createBrowserDelegate");
      }
	}

   public JBossQueue createQueue(String name) throws JMSException
   {
      try
      {
         if (closed)
         {
            throw new IllegalStateException("Session is closed");
         }
         
         CoreDestination coreDestination = cm.getCoreDestination(new JBossQueue(name));
   
         if (coreDestination == null)
         {
            throw new JMSException("There is no administratively defined queue with name:" + name);
         }
   
         return new JBossQueue(name);
      }
      catch (Throwable t)
      {
         throw ExceptionUtil.handleJMSInvocation(t, this + " createQueue");
      }
   }

   public JBossTopic createTopic(String name) throws JMSException
   {
      try
      {
         if (closed)
         {
            throw new IllegalStateException("Session is closed");
         }
         
         CoreDestination coreDestination = cm.getCoreDestination(new JBossTopic(name));
   
         if (coreDestination == null)
         {
            throw new JMSException("There is no administratively defined topic with name:" + name);
         }
   
         return new JBossTopic(name);
      }
      catch (Throwable t)
      {
         throw ExceptionUtil.handleJMSInvocation(t, this + " createTopic");
      }
   }

   public synchronized void close() throws JMSException
   {
      try
      {
         if (closed)
         {
            throw new IllegalStateException("Session is already closed");
         }
         
         if (trace) log.trace(this + " close()");
               
         // clone to avoid ConcurrentModificationException
         HashSet consumerSet = new HashSet(consumers.values());
         
         for(Iterator i = consumerSet.iterator(); i.hasNext(); )
         {
            ((ServerConsumerEndpoint)i.next()).remove();
         }           
         
         connectionEndpoint.removeSessionDelegate(sessionID);
         
         JMSDispatcher.instance.unregisterTarget(new Integer(sessionID));
         
         closed = true;
      }
      catch (Throwable t)
      {
         throw ExceptionUtil.handleJMSInvocation(t, this + " close");
      }
   }
   
   public void closing() throws JMSException
   {
      // currently does nothing
      if (trace) log.trace(this + " closing (noop)");
   }

   public boolean isClosed()
   {
      return closed;
   }
   
   public void send(JBossMessage message) throws JMSException
   {
      try
      {       
         connectionEndpoint.sendMessage(message, null);
      }
      catch (Throwable t)
      {
         throw ExceptionUtil.handleJMSInvocation(t, this + " send");
      }
   }
   
   public synchronized void acknowledgeBatch(List ackInfos) throws JMSException
   {      
      try
      {
         if (closed)
         {
            throw new javax.jms.IllegalStateException("Cannot acknowledge batch since session is closed");
         }
                  
         Iterator iter = ackInfos.iterator();
         
         while (iter.hasNext())
         {
            AckInfo ackInfo = (AckInfo)iter.next();
            
            acknowledgeInternal(ackInfo);
         }
      }
      catch (Throwable t)
      {
         throw ExceptionUtil.handleJMSInvocation(t, this + " acknowledgeBatch");
      }
   }
   
   public synchronized void acknowledge(AckInfo ackInfo) throws JMSException
   {
      try
      {
         if (closed)
         {
            throw new javax.jms.IllegalStateException("Cannot acknowledge " + ackInfo + " since session is closed");
         }
         
         acknowledgeInternal(ackInfo);      
      }
      catch (Throwable t)
      {
         throw ExceptionUtil.handleJMSInvocation(t, this + " acknowledge");
      }
   }      
        
   public synchronized void cancelDeliveries(List ackInfos) throws JMSException
   {
      try
      {
         if (closed)
         {
            //We can safely ignore any cancels since closing the session will have cancelled anything
            //remaining
            log.warn("Call to cancelDeliveries came in after session was closed");
            
            return;
         }
         
         // deliveries must be cancelled in reverse order
          
         Set consumers = new HashSet();
         
         List forDLQ = null;
         
         for (int i = ackInfos.size() - 1; i >= 0; i--)
         {
            AckInfo ack = (AckInfo)ackInfos.get(i);
            
            // We look in the global map since the message might have come from connection consumer
            ServerConsumerEndpoint consumer =
               this.connectionEndpoint.getConsumerEndpoint(ack.getConsumerID());
   
            if (consumer == null)
            {
               throw new IllegalArgumentException("Cannot find consumer id: " + ack.getConsumerID());
            }
            
            if (ack.getDeliveryCount() >= maxDeliveryAttempts)
            {
               if (forDLQ == null)
               {
                  forDLQ = new ArrayList();
               }
               
               forDLQ.add(ack);
            }
            else
            {            
               consumer.cancelDelivery(new Long(ack.getMessageID()), ack.getDeliveryCount());
            }
            
            consumers.add(consumer);
         }
         
         //Send stuff to DLQ
         
         if (forDLQ != null)
         {
            //We do this in a tx so we don't end up with the message in both the original queue
            //and the dlq if it fails half way through
            Transaction tx = tr.createTransaction();
            
            try
            {               
               for (int i = forDLQ.size() - 1; i >= 0; i--)
               {
                  AckInfo ack = (AckInfo)forDLQ.get(i);
                  
                  ServerConsumerEndpoint consumer =
                     this.connectionEndpoint.getConsumerEndpoint(ack.getConsumerID());
         
                  if (consumer == null)
                  {
                     throw new IllegalArgumentException("Cannot find consumer id: " + ack.getConsumerID());
                  }
                  
                  consumer.sendToDLQ(new Long(ack.getMessageID()), tx);                              
               }
               
               tx.commit();
            }
            catch (Throwable t)
            {
               tx.rollback();
               
               throw t;
            }
         }
         
         // need to prompt delivery for all consumers
         
         for(Iterator i = consumers.iterator(); i.hasNext(); )
         {
            ServerConsumerEndpoint consumer = (ServerConsumerEndpoint)i.next();
            consumer.promptDelivery();
         }
      }
      catch (Throwable t)
      {
         throw ExceptionUtil.handleJMSInvocation(t, this + " cancelDeliveries");
      }
   }

   public void addTemporaryDestination(JBossDestination dest) throws JMSException
   {
      try
      {
         if (closed)
         {
            throw new IllegalStateException("Session is closed");
         }
         if (!dest.isTemporary())
         {
            throw new InvalidDestinationException("Destination:" + dest + " is not a temporary destination");
         }
         connectionEndpoint.addTemporaryDestination(dest);
         
         //FIXME - Params should not be hardcoded
         long id = connectionEndpoint.getServerPeer().getNextObjectID();
         
         cm.deployTemporaryCoreDestination(dest.isQueue(), dest.getName(), id, ms, pm, mm,
                                           connectionEndpoint.getDefaultTempQueueFullSize(),
                                           connectionEndpoint.getDefaultTempQueuePageSize(),
                                           connectionEndpoint.getDefaultTempQueueDownCacheSize());
      }
      catch (Throwable t)
      {
         throw ExceptionUtil.handleJMSInvocation(t, this + " addTemporaryDestination");
      }
   }
   
   public void deleteTemporaryDestination(JBossDestination dest) throws JMSException
   {
      try
      {
         if (closed)
         {
            throw new IllegalStateException("Session is closed");
         }
   
         if (!dest.isTemporary())
         {
            throw new InvalidDestinationException("Destination:" + dest + " is not a temporary destination");
         }
         
         //It is illegal to delete a temporary destination if there any active consumers on it
         CoreDestination destination = cm.getCoreDestination(dest);
         
         if (destination == null)
         {
            throw new InvalidDestinationException("Destination:" + dest + " does not exist");         
         }
         
         if (dest instanceof JBossQueue)
         {
            if (destination.iterator().hasNext())
            {
               throw new IllegalStateException("Cannot delete temporary destination, since it has active consumer(s)");
            }
         }
         else if (dest instanceof JBossTopic)
         {
            Iterator iter = destination.iterator();
            while (iter.hasNext())
            {
               CoreSubscription sub = (CoreSubscription)iter.next();
               if (sub.iterator().hasNext())
               {
                  throw new IllegalStateException("Cannot delete temporary destination, since it has active consumer(s)");
               }
            }
         }
         
         cm.undeployTemporaryCoreDestination(dest.isQueue(), dest.getName());
         connectionEndpoint.removeTemporaryDestination(dest);
      }
      catch (Throwable t)
      {
         throw ExceptionUtil.handleJMSInvocation(t, this + " deleteTemporaryDestination");
      }
   }
   
   public void unsubscribe(String subscriptionName) throws JMSException
   {
      log.debug(this + " unsubscribing " + subscriptionName);

      try
      {
         if (closed)
         {
            throw new IllegalStateException("Session is closed");
         }
         if (subscriptionName == null)
         {
            throw new InvalidDestinationException("Destination is null");
         }
   
         String clientID = connectionEndpoint.getClientID();
   
         if (clientID == null)
         {
            throw new JMSException("null clientID on connection");
         }
   
         DurableSubscription subscription =
            cm.getDurableSubscription(clientID, subscriptionName, ms, pm, mm);
   
         if (subscription == null)
         {
            throw new InvalidDestinationException("Cannot find durable subscription with name " +
                                                  subscriptionName + " to unsubscribe");
         }

         // Section 6.11. JMS 1.1.
         // "It is erroneous for a client to delete a durable subscription while it has an active
         // TopicSubscriber for it or while a message received by it is part of a current
         // transaction or has not been acknowledged in the session."

         if (subscription.iterator().hasNext())
         {
            throw new IllegalStateException("Cannot unsubscribe durable subscription " +
                                            subscriptionName + " since it has active subscribers");
         }

         boolean removed =
            cm.removeDurableSubscription(connectionEndpoint.getClientID(), subscriptionName);
         
         if (!removed)
         {
            throw new JMSException("Failed to remove durable subscription");
         }
         
         subscription.unsubscribe();      
      }
      catch (Throwable t)
      {
         throw ExceptionUtil.handleJMSInvocation(t, this + " unsubscribe");
      }
   }
    
   // Public --------------------------------------------------------
   
   public ServerConnectionEndpoint getConnectionEndpoint()
   {
      return connectionEndpoint;
   }

   public String toString()
   {
      return "SessionEndpoint[" + sessionID + "]";
   }

   // Package protected ---------------------------------------------

   /**
    * @return a Set<Integer>
    */
   Set getConsumerEndpointIDs()
   {
      return consumers.keySet();
   }
   
   // Protected -----------------------------------------------------
   
   protected void acknowledgeInternal(AckInfo ackInfo) throws Throwable
   {
      // If the message was delivered via a connection consumer then the message needs to be acked
      // via the original consumer that was used to feed the connection consumer - which won't be
      // one of the consumers of this session. Therefore we always look in the global map of
      // consumers held in the server peer.
      ServerConsumerEndpoint consumer =
         connectionEndpoint.getConsumerEndpoint(ackInfo.getConsumerID());

      if (consumer == null)
      {
         throw new IllegalArgumentException("Cannot find consumer id: " + ackInfo.getConsumerID());
      }
      
      consumer.acknowledge(ackInfo.getMessageID());
   }
   
   protected ServerConsumerEndpoint putConsumerEndpoint(int consumerID, ServerConsumerEndpoint d)
   {
      if (trace) { log.trace(this + " caching consumer " + consumerID); }
      return (ServerConsumerEndpoint)consumers.put(new Integer(consumerID), d);
   }

   protected ServerConsumerEndpoint getConsumerEndpoint(int consumerID)
   {
      return (ServerConsumerEndpoint)consumers.get(new Integer(consumerID));
   }
   
   protected ServerConsumerEndpoint removeConsumerEndpoint(int consumerID)
   {
      if (trace) { log.trace(this + " removing consumer " + consumerID + " from cache"); }
      return (ServerConsumerEndpoint)consumers.remove(new Integer(consumerID));
   }
   
   protected ServerBrowserEndpoint putBrowserDelegate(int browserID, ServerBrowserEndpoint sbd)
   {
      return (ServerBrowserEndpoint)browsers.put(new Integer(browserID), sbd);
   }
   
   protected ServerBrowserEndpoint getBrowserDelegate(int browserID)
   {
      return (ServerBrowserEndpoint)browsers.get(new Integer(browserID));
   }
   
   protected ServerBrowserEndpoint removeBrowserDelegate(int browserID)
   {
      return (ServerBrowserEndpoint)browsers.remove(new Integer(browserID));
   }

   /**
    * Starts this session's Consumers
    */
   protected void setStarted(boolean s) throws Throwable
   {
      for(Iterator i = new HashSet(consumers.values()).iterator(); i.hasNext(); )
      {
         ServerConsumerEndpoint sce = (ServerConsumerEndpoint)i.next();
         if (s)
         {
            sce.start();
         }
         else
         {
            sce.stop();
         }
      }
   }

   // Private -------------------------------------------------------

   // Inner classes -------------------------------------------------
}
